// Code generated by 'go generate' with gen.go. DO NOT EDIT.

// SPDX-License-Identifier: Apache-2.0
// SPDX-FileCopyrightText: 2022 The Ebitengine Authors

//go:build !cgo

#include "textflag.h"

// these stubs are here because it is not possible to go:linkname directly the C functions

TEXT _pthread_attr_getstacksize(SB), NOSPLIT|NOFRAME, $0-0
	JMP purego_pthread_attr_getstacksize(SB)

TEXT _pthread_attr_destroy(SB), NOSPLIT|NOFRAME, $0-0
	JMP purego_pthread_attr_destroy(SB)
